package dev.kske.chess.board;

import java.util.Arrays;
import java.util.Iterator;
import java.util.Objects;

import dev.kske.chess.board.Piece.Color;
import dev.kske.chess.board.Piece.Type;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>Log.java</strong><br>
 * Created: <strong>09.07.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public class Log implements Iterable<MoveNode> {

	private MoveNode root, current;

	private Color		activeColor;
	private boolean[]	castlingRights;
	private Position	enPassant;
	private int			fullmoveNumber, halfmoveClock;

	public Log() { reset(); }

	/**
	 * Creates a (partially deep) copy of another {@link Log} instance which begins
	 * with the current {@link MoveNode}.
	 * 
	 * @param other          The {@link Log} instance to copy
	 * @param copyVariations If set to {@code true}, subsequent variations of the
	 *                       current {@link MoveNode} are copied with the
	 *                       {@link Log}
	 */
	public Log(Log other, boolean copyVariations) {
		enPassant		= other.enPassant;
		castlingRights	= other.castlingRights.clone();
		activeColor		= other.activeColor;
		fullmoveNumber	= other.fullmoveNumber;
		halfmoveClock	= other.halfmoveClock;

		// The new root is the current node of the copied instance
		if (!other.isEmpty()) {
			root = new MoveNode(other.current, copyVariations);
			root.setParent(null);
			current = root;
		}
	}

	@Override
	public Iterator<MoveNode> iterator() {
		return new Iterator<MoveNode>() {

			private MoveNode	current	= root;
			private boolean		hasNext	= true;

			@Override
			public boolean hasNext() { return hasNext; }

			@Override
			public MoveNode next() {
				MoveNode result = current;
				if (current.hasVariations()) current = current.getVariations().get(0);
				else hasNext = false;
				return result;
			}
		};
	}

	/**
	 * Adds a move to the move history and adjusts the log to the new position.
	 * 
	 * @param move          The move to log
	 * @param piece         The piece that performed the move
	 * @param capturedPiece The piece captured with the move
	 */
	public void add(Move move, Piece piece, Piece capturedPiece) {
		enPassant = piece.getType() == Type.PAWN && move.yDist == 2 ? new Position(move.pos.x, move.pos.y + move.ySign) : null;
		if (activeColor == Color.BLACK) ++fullmoveNumber;
		if (piece.getType() == Type.PAWN || capturedPiece != null) halfmoveClock = 0;
		else++halfmoveClock;
		activeColor = activeColor.opposite();

		// Disable castling rights if a king or a rook has been moved
		if (piece.getType() == Type.KING || piece.getType() == Type.ROOK) disableCastlingRights(piece, move.pos);

		final MoveNode leaf = new MoveNode(move, capturedPiece, castlingRights.clone(), enPassant, activeColor, fullmoveNumber, halfmoveClock);

		if (isEmpty()) {
			root	= leaf;
			current	= leaf;
		} else {
			current.addVariation(leaf);
			current = leaf;
		}
	}

	/**
	 * Removed the last move from the log and adjusts its state to the previous
	 * move.
	 */
	public void removeLast() {
		if (hasParent()) {
			current.getParent().getVariations().remove(current);
			current = current.getParent();
			update();
		} else reset();
	}

	public boolean isEmpty() { return root == null; }

	public boolean hasParent() { return !isEmpty() && current.hasParent(); }

	/**
	 * Reverts the log to its initial state corresponding to the default board
	 * position.
	 */
	public void reset() {
		root			= null;
		current			= null;
		castlingRights	= new boolean[] { true, true, true, true };
		enPassant		= null;
		activeColor		= Color.WHITE;
		fullmoveNumber	= 1;
		halfmoveClock	= 0;
	}

	/**
	 * Changes the current node to one of its children (variations).
	 * 
	 * @param index the index of the variation to select
	 */
	public void selectNextNode(int index) {
		if (!isEmpty() && current.hasVariations() && index < current.getVariations().size()) {
			current = current.getVariations().get(index);
			update();
		}
	}

	/**
	 * Selects the parent of the current {@link MoveNode} as the current node.
	 */
	public void selectPreviousNode() {
		if (hasParent()) {
			current = current.getParent();
			update();
		}
	}

	/**
	 * Selects the root {@link MoveNode} as the current node.
	 */
	public void selectRootNode() {
		if (!isEmpty()) {
			current = root;
			update();
		}
	}

	private void update() {
		activeColor		= current.activeColor;
		castlingRights	= current.castlingRights.clone();
		enPassant		= current.enPassant;
		fullmoveNumber	= current.fullmoveCounter;
		halfmoveClock	= current.halfmoveClock;
	}

	private void disableCastlingRights(Piece piece, Position initialPosition) {
		// Kingside
		if (piece.getType() == Type.KING || piece.getType() == Type.ROOK && initialPosition.x == 7)
			castlingRights[piece.getColor() == Color.WHITE ? MoveNode.WHITE_KINGSIDE : MoveNode.BLACK_KINGSIDE] = false;

		// Queenside
		if (piece.getType() == Type.KING || piece.getType() == Type.ROOK && initialPosition.x == 0)
			castlingRights[piece.getColor() == Color.WHITE ? MoveNode.WHITE_QUEENSIDE : MoveNode.BLACK_QUEENSIDE] = false;
	}

	@Override
	public int hashCode() {
		final int	prime	= 31;
		int			result	= 1;
		result	= prime * result + Arrays.hashCode(castlingRights);
		result	= prime * result + Objects.hash(activeColor, current, enPassant, fullmoveNumber, halfmoveClock);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (obj == null) return false;
		if (getClass() != obj.getClass()) return false;
		Log other = (Log) obj;
		return activeColor == other.activeColor && Arrays.equals(castlingRights, other.castlingRights) && Objects.equals(current, other.current)
				&& Objects.equals(enPassant, other.enPassant) && fullmoveNumber == other.fullmoveNumber && halfmoveClock == other.halfmoveClock;
	}

	/**
	 * @return The first logged move, or {@code null} if there is none
	 */
	public MoveNode getRoot() { return root; }

	/**
	 * @return the last logged move, or {@code null} if there is none
	 */
	public MoveNode getLast() { return current; }

	public boolean[] getCastlingRights() { return castlingRights; }

	public void setCastlingRights(boolean[] castlingRights) { this.castlingRights = castlingRights; }

	public Position getEnPassant() { return enPassant; }

	public void setEnPassant(Position enPassant) { this.enPassant = enPassant; }

	public Color getActiveColor() { return activeColor; }

	public void setActiveColor(Color activeColor) { this.activeColor = activeColor; }

	public int getFullmoveNumber() { return fullmoveNumber; }

	public void setFullmoveNumber(int fullmoveCounter) { this.fullmoveNumber = fullmoveCounter; }

	public int getHalfmoveClock() { return halfmoveClock; }

	public void setHalfmoveClock(int halfmoveClock) { this.halfmoveClock = halfmoveClock; }
}