package dev.kske.chess.ui;

import java.awt.Toolkit;
import java.awt.datatransfer.StringSelection;
import java.io.File;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;

import javax.swing.JComboBox;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;

import dev.kske.chess.board.FENString;
import dev.kske.chess.exception.ChessException;
import dev.kske.chess.game.Game;
import dev.kske.chess.io.EngineUtil;
import dev.kske.chess.pgn.PGNDatabase;
import dev.kske.chess.pgn.PGNGame;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>MenuBar.java</strong><br>
 * Created: <strong>16.07.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public class MenuBar extends JMenuBar {

	private static final long serialVersionUID = -7221583703531248228L;

	private final MainWindow mainWindow;

	public MenuBar(MainWindow mainWindow) {
		this.mainWindow = mainWindow;

		initGameMenu();
		initEngineMenu();
		initToolsMenu();
	}

	private void initGameMenu() {
		JMenu gameMenu = new JMenu("Game");

		JMenuItem newGameMenuItem = new JMenuItem("New Game");
		newGameMenuItem.addActionListener((evt) -> DialogUtil.showGameConfigurationDialog(mainWindow, (whiteName, blackName) -> {
			GamePane	gamePane	= mainWindow.addGamePane();
			Game		game		= new Game(gamePane.getBoardPane(), whiteName, blackName);
			gamePane.setGame(game);
			game.start();
		}));
		gameMenu.add(newGameMenuItem);

		JMenuItem loadFileMenu = new JMenuItem("Load game file");
		loadFileMenu.addActionListener((evt) -> DialogUtil.showFileSelectionDialog(mainWindow, this::loadFile));
		gameMenu.add(loadFileMenu);

		add(gameMenu);
		newGameMenuItem.doClick();
	}

	private void initEngineMenu() {
		JMenu engineMenu = new JMenu("Engine");

		JMenuItem addEngineMenuItem = new JMenuItem("Add engine");
		addEngineMenuItem.addActionListener((evt) -> {
			String enginePath = JOptionPane
				.showInputDialog(getParent(), "Enter the path to a UCI-compatible chess engine:", "Engine selection", JOptionPane.QUESTION_MESSAGE);
			if (enginePath != null) EngineUtil.addEngine(enginePath);
		});

		JMenuItem showInfoMenuItem = new JMenuItem("Show engine info");

		engineMenu.add(addEngineMenuItem);
		engineMenu.add(showInfoMenuItem);
		add(engineMenu);
	}

	private void initToolsMenu() {
		JMenu toolsMenu = new JMenu("Tools");

		JMenuItem exportFENMenuItem = new JMenuItem("Export board to FEN");
		exportFENMenuItem.addActionListener((evt) -> {
			final String fen = new FENString(mainWindow.getSelectedGamePane().getGame().getBoard()).toString();
			Toolkit.getDefaultToolkit().getSystemClipboard().setContents(new StringSelection(fen), null);
			JOptionPane.showMessageDialog(mainWindow, String.format("FEN-string copied to clipboard!%n%s", fen));
		});
		toolsMenu.add(exportFENMenuItem);

		JMenuItem loadFromFENMenuItem = new JMenuItem("Load board from FEN");
		loadFromFENMenuItem.addActionListener((evt) -> {
			final GamePane	gamePane	= mainWindow.addGamePane();
			final String	fen			= JOptionPane.showInputDialog("Enter a FEN string: ");
			DialogUtil.showGameConfigurationDialog(mainWindow, (whiteName, blackName) -> {
				Game game;
				try {
					game = new Game(gamePane.getBoardPane(), whiteName, blackName, new FENString(fen).getBoard());
					gamePane.setGame(game);
					game.start();
				} catch (ChessException e) {
					e.printStackTrace();
					JOptionPane
						.showMessageDialog(mainWindow, "Failed to load FEN string: " + e.toString(), "FEN loading error", JOptionPane.ERROR_MESSAGE);
				}
			});
		});
		toolsMenu.add(loadFromFENMenuItem);

		add(toolsMenu);
	}

	private void loadFile(File file) {
		final String	name		= file.getName().substring(0, file.getName().lastIndexOf('.'));
		final String	extension	= file.getName().substring(file.getName().lastIndexOf('.')).toLowerCase();
		switch (extension) {
			case ".fen":
				try {
					final GamePane	gamePane	= mainWindow.addGamePane(name);
					final String	fen			= new String(Files.readAllBytes(file.toPath()), StandardCharsets.UTF_8);
					DialogUtil.showGameConfigurationDialog(mainWindow, (whiteName, blackName) -> {
						Game game;
						try {
							game = new Game(gamePane.getBoardPane(), whiteName, blackName, new FENString(fen).getBoard());
							gamePane.setGame(game);
							game.start();
						} catch (ChessException e) {
							e.printStackTrace();
							JOptionPane.showMessageDialog(mainWindow,
									"Failed to load FEN string: " + e.toString(),
									"FEN loading error",
									JOptionPane.ERROR_MESSAGE);
						}
					});
				} catch (Exception e) {
					e.printStackTrace();
					JOptionPane.showMessageDialog(mainWindow,
							"Failed to load the file " + file.getName() + ": " + e.toString(),
							"File loading error",
							JOptionPane.ERROR_MESSAGE);
				}
				break;
			case ".pgn":
				try {
					final GamePane	gamePane	= mainWindow.addGamePane(name);
					PGNDatabase		pgnDB		= new PGNDatabase();
					pgnDB.load(file);
					if (pgnDB.getGames().size() > 0) {
						String[] gameNames = new String[pgnDB.getGames().size()];
						for (int i = 0; i < gameNames.length; i++) {
							final PGNGame game = pgnDB.getGames().get(i);
							gameNames[i] = String.format("%s vs %s: %s", game.getTag("White"), game.getTag("Black"), game.getTag("Result"));
						}
						JComboBox<String> comboBox = new JComboBox<>(gameNames);
						JOptionPane.showInputDialog(mainWindow, comboBox, "Select a game", JOptionPane.QUESTION_MESSAGE);
						DialogUtil.showGameConfigurationDialog(mainWindow, (whiteName, blackName) -> {
							final Game game = new Game(gamePane.getBoardPane(), whiteName, blackName,
									pgnDB.getGames().get(comboBox.getSelectedIndex()).getBoard());
							game.start();
						});
					}
				} catch (Exception e) {
					e.printStackTrace();
					JOptionPane.showMessageDialog(mainWindow,
							"Failed to load the file " + file.getName() + ": " + e.toString(),
							"File loading error",
							JOptionPane.ERROR_MESSAGE);
				}
				break;
			default:
				JOptionPane.showMessageDialog(mainWindow,
						"The file extension '" + extension + "' is not supported!",
						"File loading error",
						JOptionPane.ERROR_MESSAGE);
		}
	}

	public void loadFENFile(File file) {

	}
}
